/*
	Engineering Test: Boggle Solver
	Gordon Lee

*******************************************************************************

	@filename	cApplication.cpp

	@description
		Class for Application that handles the window and run loop

	@author		Gordon Lee
*/

// Includes
#include "cApplication.h"
#include "..\resource.h"
#include <io.h>
#include <fcntl.h>
#include <tchar.h>
#include <fstream>
#include <commdlg.h>
#include "cMemoryManager.h"
#include "cList.h"
#include "cBoard.h"
#include "BoggleSolver.h"

namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Inline Proxy Win Proc

	@param HWND _hWnd
	@param UINT _msg
	@param WPARAM _wParam
	@param LPARAM _lParam

	@return LRESULT CALLBACK
*/
inline LRESULT CALLBACK ProxyWndProc(HWND _hWnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
	return g_pcApplication->AppWndProc(_hWnd, _msg, _wParam, _lParam);
}

/**
	Constructor
*/
cApplication::cApplication()
: m_winClassName(""), m_hInstance(NULL), m_hWnd(NULL), m_hInputWindow(NULL), m_hOutputWindow(NULL),
	m_hOutputButton(NULL), m_hMinimumLettersComboBox(NULL), m_bConsoleOpen(false), m_dictionary(NULL)
{
}

/**
	Deconstructor
*/
cApplication::~cApplication()
{
	delete m_dictionary;
	cMemoryManager::Destroy();

	// Handle application release
	UnregisterClass(m_winClassName.c_str(), m_hInstance);
}

/**
	Instance

	@return cApplication *
*/
cApplication *cApplication::Instance(void)
{
	static cApplication *s_application = new cApplication();
	return s_application;
}

/**
	Create

	@return cApplication *
*/
cApplication *cApplication::Create(void)
{
	return Instance();
}

/**
	Destroy
*/
void cApplication::Destroy(void)
{
	delete Instance();
}

/**
	Initialize

	@param HINSTANCE _hInstance
	@param std::string _title

	@return bool
*/
bool cApplication::Initialize(HINSTANCE _hInstance, std::string _title)
{
	HWND _hWnd;
	m_hInstance = _hInstance;

	m_winClassName = "Test";

	WNDCLASSEX winClass	    = { 0 };
	winClass.lpszClassName	= m_winClassName.c_str();
	winClass.cbSize			= sizeof(WNDCLASSEX);
	winClass.style			= CS_HREDRAW | CS_VREDRAW;
	winClass.lpfnWndProc	= ProxyWndProc;
	winClass.hInstance		= _hInstance;
	winClass.lpszMenuName	= (LPCTSTR)IDR_MENU1;
	winClass.hCursor		= LoadCursor(0, IDC_ARROW);
	winClass.hIcon			= LoadIcon(_hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
	winClass.hIconSm		= LoadIcon(winClass.hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
	winClass.hbrBackground	= (HBRUSH)COLOR_WINDOW;
	if (!RegisterClassEx(&winClass))
	{
		MessageBox(NULL, _T("Call to RegisterClassEx failed!"), _T("Reverse Words - cApplication::Initialize"), NULL);
		return false;
	}
	_hWnd = CreateWindowEx(0, m_winClassName.c_str(), _title.c_str(), WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
							CW_USEDEFAULT, CW_USEDEFAULT, SCREEN_WIDTH, SCREEN_HEIGHT, GetDesktopWindow(), 0, m_hInstance, NULL);
	if (!_hWnd)
	{
		MessageBox(NULL, _T("Call to CreateWindowEx failed!"), _T("Reverse Words - cApplication::Initialize"), NULL);
		return false;
	}
	m_hWnd = _hWnd;

	// Windows and Buttons
	m_hInputWindow = CreateWindowEx(0, "EDIT", NULL,
									WS_VISIBLE | WS_CHILD | WS_BORDER | WS_VSCROLL | ES_MULTILINE,
									10, 25, 600, 450,
									m_hWnd, 0, m_hInstance, 0);

	m_hOutputWindow = CreateWindowEx(0, "EDIT", NULL,
									WS_VISIBLE | WS_CHILD | WS_BORDER | WS_VSCROLL | ES_MULTILINE | ES_READONLY,
									625, 25, 150, 450,
									m_hWnd, 0, m_hInstance, 0);

	m_hOutputButton = CreateWindowEx(0, "BUTTON", "Solve",
									WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_DEFPUSHBUTTON,
									325, 485, 150, 50,
									m_hWnd, 0, m_hInstance, 0);

	m_hMinimumLettersComboBox = CreateWindowEx(0, "COMBOBOX", NULL,
												WS_VISIBLE | WS_CHILD | WS_TABSTOP | WS_VSCROLL | CBS_DROPDOWNLIST | WS_BORDER,
												425, 0, 50, 200,
												m_hWnd, 0, m_hInstance, 0);

	SendMessage(m_hInputWindow, WM_SETFONT, (LPARAM)GetStockObject(ANSI_FIXED_FONT), 0);
	SendMessage(m_hOutputWindow, EM_LIMITTEXT, 0, 0);
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"2");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"3");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"4");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"5");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"6");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"7");
	SendMessage(m_hMinimumLettersComboBox, CB_ADDSTRING, 0, (LPARAM)"8");
	SendMessage(m_hMinimumLettersComboBox, CB_SELECTSTRING, 0, (LPARAM)"4");

	ShowWindow(m_hWnd, true);
	UpdateWindow(m_hWnd);

	// Create Memory Manager
	cMemoryManager::Create();

	// Create dictionary
	try
	{
		m_dictionary = new cDictionary(DICTIONARY_FILE);
	}
	catch (std::exception &_e)
	{
		printf("%s\n", _e.what());
		MessageBox(NULL, _T(_e.what()), _T("Warning!"), MB_ICONWARNING);
	}

	return true;
}

/**
	Open Console Window

	@return bool
*/
bool cApplication::OpenConsoleWindow(void)
{
	// Console window already open
	if (m_bConsoleOpen)
		return false;

	HANDLE _hStdHandle;
	int _hConHandle;
	FILE *_fp;

	if (!AllocConsole())
		return false;

	SetConsoleTitle("Boggle Solver Debug Console");
	_hStdHandle = GetStdHandle(STD_OUTPUT_HANDLE);
	_hConHandle = _open_osfhandle((intptr_t)_hStdHandle, _O_TEXT);
	_fp = _fdopen(_hConHandle, "w");
	*stdout = *_fp;
	setvbuf(stdout, NULL, _IONBF, 0);
	std::ios::sync_with_stdio();

	m_bConsoleOpen = true;

	return true;
}

/**
	Close Console Window

	@return bool
*/
bool cApplication::CloseConsoleWindow(void)
{
	if (!FreeConsole())
		return false;

	m_bConsoleOpen = false;

	return true;
}

/**
	Run
*/
void cApplication::Run(void)
{
	// Message loop
	MSG msg = { 0 };

	while(msg.message != WM_QUIT)
	{
        if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
}

/**
	App Wnd Proc

	@param HWND _hWnd
	@param UINT _msg
	@param WPARAM _wParam
	@param LPARAM _lParam

	@return LRESULT
*/
LRESULT cApplication::AppWndProc(HWND _hWnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
	PAINTSTRUCT _ps;
	HDC _hdc;

	switch (_msg)
	{
		case WM_CREATE:
			break;
		case WM_DESTROY:
			CloseConsoleWindow();
			PostQuitMessage(0);
			break;
		case WM_PAINT:
			_hdc = BeginPaint(m_hWnd, &_ps);		

			SetBkMode(_hdc, TRANSPARENT);
			TextOut(_hdc, 10, 10, "Board Input:", 12);
			TextOut(_hdc, 625, 10, "Valid Words:", 12);
			TextOut(_hdc, 250, 5, "Minimum Letter Words:", 21);

			EndPaint(m_hWnd, &_ps);
			break;
		case WM_COMMAND:
			switch (_wParam)
			{
			// File/New Board
			case ID_FILE_NEWBOARD:
				// Clear Input and Output
				SendMessage(m_hInputWindow, WM_SETTEXT, 0, (LPARAM)"");
				SendMessage(m_hOutputWindow, WM_SETTEXT, 0, (LPARAM) "");
				break;
			// File/Save Board
			case ID_FILE_SAVEBOARD:
			{
				// Get text length from input window
				int _bufferLen = (int)SendMessage(m_hInputWindow, WM_GETTEXTLENGTH, NULL, NULL) + 1;
				char *_buffer = new char[_bufferLen];
				
				// Get text from input window		
				SendMessage(m_hInputWindow, WM_GETTEXT, _bufferLen, (LPARAM)_buffer);

				// File path
				char _strFilePath[BUF_LEN] = "*.brd";

				// Open a save dialog to prompt for a file
				OPENFILENAME ofn	= { 0 };
				ofn.lStructSize		= sizeof( ofn );
				ofn.lpstrFile		= _strFilePath;
				ofn.nMaxFile		= BUF_LEN;
				ofn.hwndOwner		= m_hWnd;
				ofn.lpstrTitle		= "Save As Input Board File";
				ofn.lpstrFilter		= "Board Files (*.brd)\0*.brd\0All Files (*.*)\0*.*\0";
				ofn.Flags			= OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;

				if ( !GetSaveFileName( &ofn ) )
					break;

				// Write to file
				std::ofstream _ofStream(_strFilePath, std::ios::out | std::ios::binary);

				if (!_ofStream.is_open())
					break;

				_ofStream.write(_buffer, _bufferLen);
				_ofStream.close();

				// Debug Console
				if (m_bConsoleOpen)
					printf("Saved Input to Board File: %s\n", _strFilePath);

				// Cleanup
				delete [] _buffer;
				break;
			}
			// File/Load Board
			case ID_FILE_LOADBOARD:
			{
				// File path
				char _strFilePath[BUF_LEN] = "\0";

				// Open an open dialog to prompt for a file
				OPENFILENAME ofn	= { 0 };
				ofn.lStructSize		= sizeof( ofn );
				ofn.lpstrFile		= _strFilePath;
				ofn.nMaxFile		= BUF_LEN;
				ofn.hwndOwner		= m_hWnd;
				ofn.lpstrTitle		= "Open Input Board File";
				ofn.lpstrFilter		= "Board Files (*.brd)\0*.brd\0All Files (*.*)\0*.*\0";
				ofn.Flags			= OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;

				if ( !GetOpenFileName( &ofn ) )
					break;

				// Read from file
				std::ifstream _ifStream(_strFilePath, std::ios::in | std::ios::binary | std::ios::ate);

				if (!_ifStream.is_open())
					break;

				std::ifstream::pos_type _size = _ifStream.tellg();
				size_t _bufferLen = (size_t)_size.seekpos() + 1;
				char *_buffer = new char[_bufferLen];
				_ifStream.seekg(NULL, std::ios::beg);
				_ifStream.read(_buffer, _size);
				_ifStream.close();

				memcpy_s(_buffer + _bufferLen - 1, _bufferLen, "\0", 1);

				// Set input window text
				SendMessage(m_hInputWindow, WM_SETTEXT, _bufferLen, (LPARAM)_buffer);

				// Clear Output
				SendMessage(m_hOutputWindow, WM_SETTEXT, 0, (LPARAM) "");

				// Debug Console
				if (m_bConsoleOpen)
					printf("Loaded Board Input from File: %s\n", _strFilePath);

				// Cleanup
				delete [] _buffer;
				break;
			}
			// File/Open Debug Console
			case ID_FILE_OPENDEBUGCONSOLE:
				OpenConsoleWindow();
				break;
			// File/Close Debug Console
			case ID_FILE_CLOSEDEBUGCONSOLE:
				CloseConsoleWindow();
				break;
			// File/Exit
			case ID_FILE_EXIT:
				CloseConsoleWindow();
				PostQuitMessage(0);
				break;
			// Help/How To
			case ID_HELP_HOWTO:
				MessageBox(NULL,
					_T("Menu:\nFile -> New Board - Clears Input and Output windows\nFile -> Save Board - Saves text in Input window to a .brd file\nFile -> Load Board - Loads .brd file into Input window\nFile -> Open Debug Console - Opens a console window for debug information such as algorithm time\nFile -> Close Debug Console - Closes the debug console window\n\nInstructions:\nUpon opening the application, dictionary.txt must be in the same folder as the application or else the dictionary will be missing.\nType text into Input window or Load Board from a .brd file.\nTo construct a valid board, letters must be separated by spaces, and rows must be separated by a new line.\nExample:\na b c\nd e f\ng h i\nTest files are provided, such as Test1.txt is the example problem.\nIncluded are some stress test files, Test2 is a random letter 8x8 board.\nThere is a drop down combo box for the minimum letters for words to search for. The example problem used a minimum of 4 letters, but Boggle rules accept words of minimum 3 letters."),
					_T("How To"),
					MB_ICONINFORMATION);
				break;
			// Help/About
			case ID_HELP_ABOUT:
				MessageBox(NULL,
					_T("Boggle Solver Engineer Test\nAuthor: Gordon Lee\nWritten in C++ and Microsoft Visual Studio 2008\n\nThe dictionary is the Scrabble Official Tournament Word List with each word on a separate line.\nI loaded the dictionary into a trie (prefix tree) graph for fast searching and prefixed letters can be shared. For example, there are thousands of words that start with the letter 'a' but if none exist on the board, then the algorithm searches the board for 'a' once, then skips all words starting with 'a'.\nOptimally, I would have implemented a directed acyclic word graph (DAWG) which would drasticly decrease the amount of nodes because of more shared nodes. However, due to production time constraints and project time management, I opted to stay with the trie implementation.\nBuilding the trie dictionary and searching the dictionary for the solver are both recursive because of the way the trie is setup with single child pointers and sibling pointers as the breadth of the trie.\nThe solving algorithm recursively searches the dictionary trie letter by letter forming words, while checking the board's letters within the rules of Boggle. As words are found or not found, the recursive search unravels and undos itself, as the search uses shared info.\nException handling is used to protect against a missing dictionary file and invalid board input.\nI implemented a pool-based memory manager for just the nodes in the trie because there are hundreds of thousands of nodes. Using simple new and delete would take too long so I overloaded the node's new and delete operators to work inside the pool which afterwards gets deleted in bulk chunks."),
					_T("About"),
					MB_ICONINFORMATION);
				break;
			// Button
			case BN_CLICKED:
				if ((HWND)_lParam == m_hOutputButton)
				{
					// Get text length from input window
					int _bufferLen = (int)SendMessage(m_hInputWindow, WM_GETTEXTLENGTH, NULL, NULL) + 1;
					char *_buffer = new char[_bufferLen];
					cList<std::string> *_validWords = NULL;
					int _minLetters = 0;
					
					// Get text from input window		
					SendMessage(m_hInputWindow, WM_GETTEXT, _bufferLen, (LPARAM)_buffer);

					// Create board
					cBoard _board;
					try
					{
						_board.ParseBoardInput(_buffer);
					}
					catch (std::exception &_e)
					{
						printf("%s\n", _e.what());
						MessageBox(NULL, _T(_e.what()), _T("Warning!"), MB_ICONWARNING);
						break;
					}

					// Set minimum letters
					_minLetters = SendMessage(m_hMinimumLettersComboBox, CB_GETCURSEL, 0, 0) + 2;
					m_dictionary->SetMinLetters(_minLetters);

					// Debug Console
					if (m_bConsoleOpen)
					{
						printf("Board dimensions: %dx%d\n", _board.Rows(), _board.Cols());
						printf("Minimum letter words: %d\n", m_dictionary->MinLetters());
						m_debugTimer.Start();

						// Boggle Solver
						_validWords = BoggleSolver(&_board, m_dictionary);

						unsigned long _time = m_debugTimer.GetTime();
						printf("Boggle Solver Algorithm Time: %d milliseconds\n", _time);
						printf("Number of found words: %d\n", _validWords->Size());
					}
					else
						_validWords = BoggleSolver(&_board, m_dictionary);

					// Set text to output window
					SendMessage(m_hOutputWindow, WM_SETTEXT, 0, (LPARAM) "");
					while (!_validWords->Empty())
					{
						SendMessage(m_hOutputWindow, EM_REPLACESEL, 0, (LPARAM)_validWords->Front().c_str());
						if (_validWords->Size() > 1)
							SendMessage(m_hOutputWindow, EM_REPLACESEL, 0, (LPARAM)"\r\n");
						_validWords->Pop_Front();
					}

					// Refresh window
					InvalidateRect(m_hWnd, NULL, TRUE);
					UpdateWindow(m_hWnd);

					// Cleanup
					if (_validWords)
						delete _validWords;
					delete [] _buffer;
				}
				break;
			default:
				break;
			}
		default:
			return DefWindowProc(_hWnd, _msg, _wParam, _lParam);
	}
	return DefWindowProc(_hWnd, _msg, _wParam, _lParam);
}
////////////////////////////////////////////////////////////////////////////////
}
